import com.jme.app.SimpleGame;
import com.jme.scene.Node;
import com.jme.scene.TriMesh;
import com.jme.scene.TriMesh.Mode;
import com.jme.scene.Line;
import com.jme.scene.Point;
import com.jme.scene.Text;
import com.jme.scene.TexCoords;
import com.jme.scene.state.RenderState;
import com.jme.scene.state.LightState;
import com.jme.scene.state.CullState;
import com.jme.scene.state.MaterialState;
import com.jme.scene.state.MaterialState.ColorMaterial;
import com.jme.scene.state.TextureState;
import com.jme.scene.state.BlendState;
import com.jme.scene.Spatial;
import com.jme.scene.SharedNode;
import com.jme.system.DisplaySystem;

import com.jme.renderer.Renderer;
import com.jme.renderer.ColorRGBA;

import com.jme.light.DirectionalLight;

import com.jme.image.Image;
import com.jme.image.Texture;
import com.jme.image.Texture2D;
import com.jme.image.Texture.MinificationFilter;
import com.jme.image.Texture.MagnificationFilter;

import com.jme.util.TextureManager;

import java.awt.Font;

import com.jmex.font3d.Text3D;
import com.jmex.font3d.Font3D;

import com.jme.util.geom.NormalGenerator;
import com.jme.util.geom.BufferUtils;

import com.jme.math.Vector3f;
import com.jme.math.Matrix4f;
import com.jme.math.TransformMatrix;
import com.jme.math.Quaternion;
 
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

// draws a X3DV file with IndexedTriangleSet and TriangleSet nodes

class MaterialToJme {
   static void setMaterial(MaterialState materialState, X3dNode node) {
      if (((X3dShape)node.m_parent) == null)
          return;
      X3dAppearance appearance = (X3dAppearance)((X3dShape)node.m_parent).appearance;
      if (appearance != null) {
          X3dMaterial material = (X3dMaterial)((X3dAppearance)appearance).material;

          if (material != null) {
              materialState.setAmbient(new ColorRGBA(material.diffuseColor[0],
                                                     material.diffuseColor[1],  
                                                     material.diffuseColor[2],  
                                                     1 - material.transparency)); 

              materialState.setDiffuse(new ColorRGBA(material.diffuseColor[0],
                                                     material.diffuseColor[1],
                                                     material.diffuseColor[2],
                                                     1 - material.transparency));

              materialState.setEmissive(new ColorRGBA(material.emissiveColor[0],
                                                      material.emissiveColor[1],
                                                      material.emissiveColor[2],
                                                      1 - material.transparency));
              materialState.setSpecular(new ColorRGBA(material.specularColor[0],
                                                      material.specularColor[1],
                                                      material.specularColor[2],
                                                      1 - material.transparency));
              materialState.setShininess(material.shininess * 128.0f);
              materialState.setEnabled(true);
          }
      }
   }

   static MaterialState makeMaterial(X3dNode node) {
      MaterialState materialState = DisplaySystem.getDisplaySystem().getRenderer().createMaterialState();
      setMaterial(materialState, node);
      return materialState;
   }
}

class TextureToJme {
   static TextureState makeTexture(X3dNode node) {
      if (((X3dShape)node.m_parent) == null)
          return null;
      X3dAppearance appearance = (X3dAppearance)((X3dShape)node.m_parent).appearance;
      if (appearance != null) {
          if ((appearance.texture != null) && 
              (appearance.texture.getClass().getName().equals("X3dImageTexture"))
             ) {
              float scale[] = new float[2];
              scale[0] = 1.0f;
              scale[1] = 1.0f;
              float center[] = new float[2];
              center[0] = 1.0f;
              center[1] = 1.0f;
              float rotation = 0.0f;
              float translation[] = new float[2];
              translation[0] = 1.0f;
              translation[1] = 1.0f;
              if ((appearance.textureTransform != null) && 
                  (appearance.textureTransform.getClass().getName().equals(
                   "X3dTextureTransform"))) {
                   X3dTextureTransform x3dTextureTransform = (X3dTextureTransform)appearance.textureTransform;
                   scale[0] = x3dTextureTransform.scale[0]; 
                   scale[1] = x3dTextureTransform.scale[1]; 
                   center[0] = x3dTextureTransform.center[0]; 
                   center[1] = x3dTextureTransform.center[1]; 
                   rotation = x3dTextureTransform.rotation; 
                   translation[0] = x3dTextureTransform.translation[0]; 
                   translation[1] = x3dTextureTransform.translation[1]; 
              }
              X3dImageTexture x3dTexture = (X3dImageTexture)appearance.texture;
              if (x3dTexture.url.length > 0)
                  for (int i = 0; i < x3dTexture.url.length; i++) {
                      Texture texture = TextureManager.loadTexture(x3dTexture.url[i], 
                           Texture.MinificationFilter.NearestNeighborNoMipMaps,
                           Texture.MagnificationFilter.NearestNeighbor,
                           1.0f,
                           true);
                      if (texture == null)
                          continue;
                      Matrix4f transformMatrix = new Matrix4f();
                      transformMatrix.setIdentity();
                      Matrix4f multMatrix = new Matrix4f();

                      multMatrix.setIdentity();
                      multMatrix.setTranslation(-center[0], -center[1], 0.0f);
                      transformMatrix = transformMatrix.mult(multMatrix);

                      multMatrix.setIdentity();
                      multMatrix.scale(new Vector3f(scale[0], scale[1], 0));
                      transformMatrix = transformMatrix.mult(multMatrix);
 
                      multMatrix.setIdentity();
                      multMatrix.fromAngleAxis(rotation, 
                                               new Vector3f(0.0f, 0.0f, 1.0f));
                      transformMatrix = transformMatrix.mult(multMatrix);

                      multMatrix.setIdentity();
                      multMatrix.setTranslation(center[0], center[1], 0.0f);
                      transformMatrix = transformMatrix.mult(multMatrix);

                      multMatrix.setIdentity();
                      multMatrix.setTranslation(translation[0], translation[1], 0.0f);
                      transformMatrix = transformMatrix.mult(multMatrix);

                      texture.setMatrix(transformMatrix);
                      texture.setWrap(Texture.WrapAxis.S, x3dTexture.repeatS ? Texture.WrapMode.Repeat : Texture.WrapMode.Clamp);
                      texture.setWrap(Texture.WrapAxis.T, x3dTexture.repeatT ? Texture.WrapMode.Repeat : Texture.WrapMode.Clamp);
                      TextureState textureState = DisplaySystem.getDisplaySystem().getRenderer().createTextureState();
                      textureState.setTexture(texture, 0);
                      return textureState;
                  }
          }          
      }
      return null;
   }
}

class AlphaToJme {
    static BlendState makeAlpha() {
        BlendState blendState = DisplaySystem.getDisplaySystem().getRenderer().createBlendState();
        blendState.setTestEnabled(true);
	blendState.setBlendEnabled(true);
	blendState.setSourceFunction(BlendState.SourceFunction.SourceAlpha);
	blendState.setDestinationFunction(BlendState.DestinationFunction.OneMinusDestinationAlpha);
	blendState.setTestFunction(BlendState.TestFunction.NotEqualTo);
        return blendState;
    }
}

class ColorToJme {
   static FloatBuffer makeColor(X3dNode node, int needed_length) {

      int numDefinedColors = 0;
      float r = 0.8f;
      float b = 0.8f;
      float g = 0.8f;
      float a = 1;
      float color[] = new float[needed_length * 4];
      if (node != null) {
         if (node.getClass().getName().equals("X3dColor")) {
            X3dColor colors = (X3dColor) node;
            int colorLength = colors.color.length / 3 > needed_length ? 
                              needed_length : colors.color.length / 3;
            if (colors.color.length == 3) {
                r = colors.color[0];       
                b = colors.color[1];       
                g = colors.color[2];       
            } else {      
               for (int i = 0; i < colorLength; i++) {
                  for (int j = 0; j < 3; j++)
                     color[i * 4 + j] = colors.color[i * 3 + j];
                  color[i * 4 + 3] = 1;
                  numDefinedColors++;
               }
            }           
         } else if (node.getClass().getName().equals("X3dColorRGBA")) {
             X3dColorRGBA colors = (X3dColorRGBA) node;
             for (int i = 0; i < colors.color.length; i++) {
                color[i] = colors.color[i];
                numDefinedColors++;
             }
         }
      }
      for (int i = 0; i < needed_length; i++)
         color[i] = 0.8f;
      FloatBuffer colorsBuffer = BufferUtils.createFloatBuffer(color.length);
      colorsBuffer.put(color); 
      return colorsBuffer;
   }
}

class MyTriangleSetCallback extends X3dTriangleSetDoWithDataCallback {

   public void doWithData(X3dNode node) {

      X3dTriangleSet triangles = (X3dTriangleSet) node; 

      float vertices[] = ((X3dCoordinate)triangles.coord).point;
      FloatBuffer verticesBuffer = BufferUtils.createFloatBuffer(vertices.length);
      verticesBuffer.put(vertices);

      FloatBuffer normalsBuffer = null;
      if (((X3dNormal)triangles.normal) != null) {
         float normals[] = ((X3dNormal)triangles.normal).vector;
         normalsBuffer = BufferUtils.createFloatBuffer(normals.length);
         normalsBuffer.put(normals);
      }

      FloatBuffer colorsBuffer = ColorToJme.makeColor(triangles.color, vertices.length / 3);

      FloatBuffer texCoordsBuffer = null;
      if (triangles.texCoord != null) {
         float texCoords[] = ((X3dTextureCoordinate)triangles.texCoord).point;

         if (texCoords != null) {
            texCoordsBuffer = BufferUtils.createFloatBuffer(texCoords.length);
            texCoordsBuffer.put(texCoords);
         }
      }
      int indices[] = new int[vertices.length / 3];
      for (int i = 0; i < indices.length; i++)
          indices[i] = i;
      IntBuffer indicesBuffer = BufferUtils.createIntBuffer(indices.length);
      indicesBuffer.put(indices);

      TriMesh tri = new TriMesh("TriangleSet", verticesBuffer, 
                                               normalsBuffer,
                                               colorsBuffer, 
                                               texCoordsBuffer == null ? null :
                                               new TexCoords(texCoordsBuffer),
                                               indicesBuffer);

      TextureState texture = TextureToJme.makeTexture(triangles);
      if (texture != null)
          tri.setRenderState(texture);
      if (triangles.color != null)
         tri.setLightCombineMode(Spatial.LightCombineMode.Off);
      MaterialState materialState = DisplaySystem.getDisplaySystem().getRenderer().createMaterialState();
      CullState cullState = DisplaySystem.getDisplaySystem().getRenderer().createCullState();
      materialState.setMaterialFace(MaterialState.MaterialFace.FrontAndBack);
      if (triangles.solid)
          if (triangles.ccw)
              cullState.setCullFace(CullState.Face.Back);
          else
              cullState.setCullFace(CullState.Face.Front);         
      MaterialToJme.setMaterial(materialState, triangles);
      tri.clearRenderState(RenderState.StateType.Material);
      tri.setRenderState(materialState);
      tri.setRenderState(cullState);

      node.m_data = tri;             
   }
}

class MyIndexedTriangleSetCallback extends X3dIndexedTriangleSetDoWithDataCallback {

   public void doWithData(X3dNode node) {

      X3dIndexedTriangleSet triangles = (X3dIndexedTriangleSet) node; 

      float vertices[] = ((X3dCoordinate)triangles.coord).point;
      FloatBuffer verticesBuffer = BufferUtils.createFloatBuffer(vertices.length);
      verticesBuffer.put(vertices);

      FloatBuffer normalsBuffer = null;
      if (((X3dNormal)triangles.normal) != null) {
         float normals[] = ((X3dNormal)triangles.normal).vector;
         normalsBuffer = BufferUtils.createFloatBuffer(normals.length);
         normalsBuffer.put(normals);
      }

      FloatBuffer colorsBuffer = ColorToJme.makeColor(triangles.color, vertices.length / 3);

      FloatBuffer texCoordsBuffer = null;
      if (triangles.texCoord != null) {
         float texCoords[] = ((X3dTextureCoordinate)triangles.texCoord).point;

         if (texCoords != null) {
            texCoordsBuffer = BufferUtils.createFloatBuffer(texCoords.length);
            texCoordsBuffer.put(texCoords);
         }
      }
      int indices[] = triangles.index;
      IntBuffer indicesBuffer = BufferUtils.createIntBuffer(indices.length);
      indicesBuffer.put(indices);

      TriMesh tri = new TriMesh("IndexedTriangleSet", 
                                verticesBuffer, 
                                normalsBuffer,
                                colorsBuffer, 
//                                texCoordsBuffer == null ? null :
                                new TexCoords(texCoordsBuffer),
                                indicesBuffer);
      TextureState texture = TextureToJme.makeTexture(triangles);
      if (texture != null)
          tri.setRenderState(texture);
      if (triangles.color != null)
         tri.setLightCombineMode(Spatial.LightCombineMode.Off);
      MaterialState materialState = DisplaySystem.getDisplaySystem().getRenderer().createMaterialState();
      CullState cullState = DisplaySystem.getDisplaySystem().getRenderer().createCullState();
      materialState.setMaterialFace(MaterialState.MaterialFace.FrontAndBack);
      if (triangles.solid)
          if (triangles.ccw)
              cullState.setCullFace(CullState.Face.Back);
          else
              cullState.setCullFace(CullState.Face.Front);         
      MaterialToJme.setMaterial(materialState, triangles);
      tri.clearRenderState(RenderState.StateType.Material);
      tri.setRenderState(materialState);
      tri.setRenderState(cullState);

      node.m_data = tri;             
   }
}

class MyIndexedLineSetCallback extends X3dIndexedLineSetDoWithDataCallback {

   public void doWithData(X3dNode node) {

      X3dIndexedLineSet lines = (X3dIndexedLineSet) node; 

      if (lines.coord == null)
          return;

      int indices[] = lines.coordIndex;
      float vertices[] = ((X3dCoordinate)lines.coord).point;
      float growingVertices[] = new float[indices.length * 2 * 3];
      int growingVerticesLen = 0;
      for (int i = 1; i < indices.length; i++) {
          if ((indices[i] > -1) && (indices[i - 1] > -1))  {
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i - 1]];
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i - 1] + 1];
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i - 1] + 2];
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i]];
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i] + 1];
             growingVertices[growingVerticesLen++] = vertices[3 * indices[i] + 2];
          } 
      }
      float lineVertices[] = new float[growingVerticesLen];
      for (int i = 0; i < growingVerticesLen; i++)
         lineVertices[i] = growingVertices[i];      
      FloatBuffer verticesBuffer = BufferUtils.createFloatBuffer(lineVertices.length);
      verticesBuffer.put(lineVertices);

      FloatBuffer colorsBuffer = ColorToJme.makeColor(lines.color, lineVertices.length);

      FloatBuffer normalsBuffer = null;
      TexCoords texCoords = new TexCoords(null);
      Line line = new Line("IndexedLineSet", verticesBuffer, 
                                             normalsBuffer,
                                             colorsBuffer, 
                                             texCoords);
      line.setRenderState(MaterialToJme.makeMaterial(lines));
      TextureState texture = TextureToJme.makeTexture(lines);
      if (texture != null)
          line.setRenderState(texture);
      line.setRenderState(AlphaToJme.makeAlpha());
      if (lines.color != null)
         line.setLightCombineMode(Spatial.LightCombineMode.Off);
       
      node.m_data = line;             
   }
}

class MyPointSetCallback extends X3dPointSetDoWithDataCallback {

   public void doWithData(X3dNode node) {

      X3dPointSet points = (X3dPointSet) node; 

      if (points.coord == null)
          return;

      float vertices[] = ((X3dCoordinate)points.coord).point;
      FloatBuffer verticesBuffer = BufferUtils.createFloatBuffer(vertices.length);
      verticesBuffer.put(vertices);

      FloatBuffer colorsBuffer = ColorToJme.makeColor(points.color, vertices.length);

      FloatBuffer normalsBuffer = null;
      TexCoords texCoords = new TexCoords(null);
      Point point = new Point("PointSet", verticesBuffer, 
                                             normalsBuffer,
                                             colorsBuffer, 
                                             texCoords);
      point.setRenderState(MaterialToJme.makeMaterial(points));
      TextureState texture = TextureToJme.makeTexture(points);
      if (texture != null)
          point.setRenderState(texture);
      point.setRenderState(AlphaToJme.makeAlpha());
      if (points.color != null)
         point.setLightCombineMode(Spatial.LightCombineMode.Off);
       
      node.m_data = point;             
   }
}

class MyTextCallback extends X3dTextDoWithDataCallback {

   public void doWithData(X3dNode node) {

      X3dText text = (X3dText) node; 

      if ((text.string != null) && (text.string.length > 0)) {
      
          X3dFontStyle fontStyle = (X3dFontStyle)text.fontStyle;
          int size = 1;
          String family = "Serif";
          int style = Font.PLAIN;
          Font font = null;
          if (fontStyle != null) {
              size = fontStyle.size < 1 ? 1 : (int)fontStyle.size;
              if (fontStyle.style != null) {
                  if (fontStyle.style.equals("PLAIN"))
                      style = Font.PLAIN;
                  else if (fontStyle.style.equals("BOLD"))
                      style = Font.BOLD;
                  else if (fontStyle.style.equals("ITALIC"))
                      style = Font.ITALIC;
                  else if (fontStyle.style.equals("BOLDITALIC"))
                      style = Font.ITALIC | Font.BOLD;
                  else
                      style = Font.PLAIN;
              }
              if (fontStyle.family != null) {
                  for (int i = 0; i < fontStyle.family.length; i++) {
                      if (fontStyle.family[i].equals("SERIF")) {
                          family = "Serif";
                      } else if (fontStyle.family[i].equals("SANS")) {
                          family = "SansSerif";
                      } else if (fontStyle.family[i].equals("TYPEWRITER")) {
                          family = "Monospaced";
                      } else {
                          font = new Font(fontStyle.family[i], style, size);
                          if (font.getFamily().equals(fontStyle.family[i]))
                              break;
                          else
                              continue;
                      }
                      font = new Font(family, style, size);
                      break;
                  }
              } else
                  font = new Font(family, style, size);              
          } else
              font = new Font(family, style, size);              
          Text3D txt = null;
          try {
              Font3D font3d = new Font3D(font, 1, false, true, false);
              txt = font3d.createText(text.string[0], size, style);
          } catch (NullPointerException e) {
              e.printStackTrace();
              System.err.println("Strange Exception catched, trying again with default font Serif"); 
              font = new Font("Serif", style, size);              
              Font3D font3d = new Font3D(font, 1, false, true, false);
              txt = font3d.createText(text.string[0], size, style);
          }
          txt.setLocalScale(new Vector3f(1, 1, 0.001f));
          txt.setRenderState(MaterialToJme.makeMaterial(text));

          node.m_data = txt;             
      }
   }
}

class MyShapeCallback extends X3dShapeDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dShape shape = (X3dShape) node;
      Node jmeShape = new Node("Shape");
      if (shape.geometry != null)
         if (shape.geometry.m_data != null)
            jmeShape.attachChild(((Spatial)shape.geometry.m_data));
      node.m_data = jmeShape;       
   }
}

class MyTransformCallback extends X3dTransformDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dTransform transform = (X3dTransform) node;

      Node jmeTransform = new Node("Transform");
      TransformMatrix matrix = new TransformMatrix();
      TransformMatrix multMatrix = new TransformMatrix();
      Vector3f tempStore = new Vector3f();
      Quaternion quat = new Quaternion();

      multMatrix.setTranslation(transform.translation[0], 
                                transform.translation[1],
                                transform.translation[2]);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setTranslation(transform.center[0], 
                                transform.center[1],
                                transform.center[2]);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      quat.fromAngleAxis(transform.rotation[3], 
                         new Vector3f(transform.rotation[0],
                                      transform.rotation[1],
                                      transform.rotation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      // scaleOrientation (shearing ?) not supported by TransformMatrix ?
      multMatrix.loadIdentity();
      quat.loadIdentity(); 
      quat.fromAngleAxis(transform.scaleOrientation[3], 
                         new Vector3f(transform.scaleOrientation[0],
                                      transform.scaleOrientation[1],
                                      transform.scaleOrientation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setScale(new Vector3f(transform.scale[0], 
                                       transform.scale[1],
                                       transform.scale[2]));
      matrix.multLocal(multMatrix, tempStore);

      // scaleOrientation (shearing ?) not supported by TransformMatrix ?
      multMatrix.loadIdentity();
      quat.loadIdentity(); 
      quat.fromAngleAxis(-transform.scaleOrientation[3], 
                         new Vector3f(transform.scaleOrientation[0],
                                      transform.scaleOrientation[1],
                                      transform.scaleOrientation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setTranslation(-transform.center[0], 
                                -transform.center[1],
                                -transform.center[2]);
      matrix.multLocal(multMatrix, tempStore);

      matrix.applyToSpatial(jmeTransform);
      if (transform.children != null) {
         for (int i = 0; i < transform.children.length; i++) {
            if (transform.children[i] == null)
               continue;
            if (transform.children[i].m_data == null)
               continue;
            jmeTransform.attachChild(((Node)transform.children[i].m_data));
         }
      }
      node.m_data = jmeTransform;       
   }
}

class MyGroupCallback extends X3dGroupDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dGroup group = (X3dGroup)node;
      Node jmeGroup = new Node("Group");
      if (node.m_parent != null) // not root node
         node.m_data = jmeGroup;       
      if (group.children != null) {
         for (int i = 0; i < group.children.length; i++) {
            if (group.children[i] == null)
               continue;
            if (group.children[i].m_data == null)
               continue;
            ((Node)node.m_data).attachChild(((Node)group.children[i].m_data));
         }
      }
   }
}

class MyAnchorCallback extends X3dAnchorDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dAnchor anchor = (X3dAnchor)node;
      Node jmeAnchor = new Node("Anchor");
      if (anchor.children != null) {
         for (int i = 0; i < anchor.children.length; i++) {
            if (anchor.children[i] == null)
               continue;
            if (anchor.children[i].m_data == null)
               continue;
            jmeAnchor.attachChild(((Node)anchor.children[i].m_data));
         }
      }
      node.m_data = jmeAnchor;       
   }
}

class MySwitchCallback extends X3dSwitchDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dSwitch swith = (X3dSwitch)node;
      if ((swith.whichChoice < 0) || 
          (swith.whichChoice >= swith.children.length))
         return;
      Node jmeSwitch = new Node("Switch");
      if ((swith.children[swith.whichChoice] != null) && 
          (swith.children[swith.whichChoice].m_data != null))
         jmeSwitch.attachChild(((Node)swith.children[swith.whichChoice].m_data));
      node.m_data = jmeSwitch;       
   }
}

class MyCollisionCallback extends X3dCollisionDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dCollision collision = (X3dCollision)node;
      Node jmeCollision = new Node("Collision");
      if (collision.children != null) {
         for (int i = 0; i < collision.children.length; i++) {
            if (collision.children[i] == null)
               continue;
            if (collision.children[i].m_data == null)
               continue;
            jmeCollision.attachChild(((Node)collision.children[i].m_data));
         }
      }
      node.m_data = jmeCollision;       
   }
}

class MyCADPartCallback extends X3dCADPartDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dCADPart cadPart = (X3dCADPart)node;
      Node jmeCADPart = new Node("CADPart");

      TransformMatrix matrix = new TransformMatrix();
      TransformMatrix multMatrix = new TransformMatrix();
      Vector3f tempStore = new Vector3f();
      Quaternion quat = new Quaternion();

      multMatrix.setTranslation(cadPart.translation[0], 
                                cadPart.translation[1],
                                cadPart.translation[2]);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setTranslation(cadPart.center[0], 
                                cadPart.center[1],
                                cadPart.center[2]);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      quat.fromAngleAxis(cadPart.rotation[3], 
                         new Vector3f(cadPart.rotation[0],
                                      cadPart.rotation[1],
                                      cadPart.rotation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      // scaleOrientation (shearing ?) not supported by TransformMatrix ?
      multMatrix.loadIdentity();
      quat.loadIdentity(); 
      quat.fromAngleAxis(cadPart.scaleOrientation[3], 
                         new Vector3f(cadPart.scaleOrientation[0],
                                      cadPart.scaleOrientation[1],
                                      cadPart.scaleOrientation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setScale(new Vector3f(cadPart.scale[0], 
                                       cadPart.scale[1],
                                       cadPart.scale[2]));
      matrix.multLocal(multMatrix, tempStore);

      // scaleOrientation (shearing ?) not supported by TransformMatrix ?
      multMatrix.loadIdentity();
      quat.loadIdentity(); 
      quat.fromAngleAxis(-cadPart.scaleOrientation[3], 
                         new Vector3f(cadPart.scaleOrientation[0],
                                      cadPart.scaleOrientation[1],
                                      cadPart.scaleOrientation[2]));
      multMatrix.setRotationQuaternion(quat);
      matrix.multLocal(multMatrix, tempStore);

      multMatrix.loadIdentity();
      multMatrix.setTranslation(-cadPart.center[0], 
                                -cadPart.center[1],
                                -cadPart.center[2]);
      matrix.multLocal(multMatrix, tempStore);

      matrix.applyToSpatial(jmeCADPart);
      if (cadPart.children != null) {
         for (int i = 0; i < cadPart.children.length; i++) {
            if (cadPart.children[i] == null)
               continue;
            if (cadPart.children[i].m_data == null)
               continue;
            jmeCADPart.attachChild(((Node)cadPart.children[i].m_data));
         }
      }
      node.m_data = jmeCADPart;       
   }
}

class MyCADAssemblyCallback extends X3dCADAssemblyDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dCADAssembly cadAssembly = (X3dCADAssembly)node;
      Node jmeCADAssembly = new Node("CADAssembly");
      if (cadAssembly.children != null) {
         for (int i = 0; i < cadAssembly.children.length; i++) {
            if (cadAssembly.children[i] == null)
               continue;
            if (cadAssembly.children[i].m_data == null)
               continue;
            jmeCADAssembly.attachChild(((Node)cadAssembly.children[i].m_data));
         }
      }
      node.m_data = jmeCADAssembly;       
   }
}

class MyCADLayerCallback extends X3dCADLayerDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dCADLayer cadLayer = (X3dCADLayer)node;
      Node jmeCADLayer = new Node("CADLayer");
      node.m_data = jmeCADLayer;       
      if (cadLayer.children == null)
          return;
      for (int i = 0; i < cadLayer.children.length; i++) {
         if (cadLayer.children[i] == null)
            continue;
         if (cadLayer.children[i].m_data == null)
            continue;
         if ((cadLayer.visible == null) || (i >= cadLayer.visible.length))
            ((Node)node.m_data).attachChild(((Node)cadLayer.children[i].m_data));
         else if (cadLayer.visible[i])
            ((Node)node.m_data).attachChild(((Node)cadLayer.children[i].m_data));
      }
   }
}

class MyCADFaceCallback extends X3dCADFaceDoWithDataCallback {

   public void doWithData(X3dNode node) {
      X3dCADFace cadFace = (X3dCADFace)node;
      Node jmeCADFace = new Node("CADFace");
      node.m_data = jmeCADFace;       
      ((Node)node.m_data).attachChild(((Node)cadFace.shape.m_data));
   }
}

public class main extends SimpleGame{

   public static void main(String args[]) {
      main app = new main();
      app.start();
   }                                             
                                                          
   protected void simpleInitGame(){              
      X3dSceneGraph sceneGraph = new X3dSceneGraph();
      MyTriangleSetCallback myTriangleSet = new MyTriangleSetCallback();
      MyIndexedTriangleSetCallback myIndexedTriangleSet = new MyIndexedTriangleSetCallback();
      MyTextCallback myText = new MyTextCallback();
      MyIndexedLineSetCallback myIndexedLineSet = new MyIndexedLineSetCallback();
      MyPointSetCallback myPointSet = new MyPointSetCallback();
      MyShapeCallback myShape = new MyShapeCallback();
      MyTransformCallback myTransform = new MyTransformCallback();
      MyGroupCallback myGroup = new MyGroupCallback();
      MyAnchorCallback myAnchor = new MyAnchorCallback();
      MySwitchCallback mySwitch = new MySwitchCallback();
      MyCollisionCallback myCollision = new MyCollisionCallback();
      MyCADPartCallback myCADPart = new MyCADPartCallback();
      MyCADAssemblyCallback myCADAssembly = new MyCADAssemblyCallback();
      MyCADLayerCallback myCADLayer = new MyCADLayerCallback();
      MyCADFaceCallback myCADFace = new MyCADFaceCallback();
      sceneGraph.setX3dTriangleSetDoWithDataCallback(myTriangleSet);
      sceneGraph.setX3dIndexedTriangleSetDoWithDataCallback(myIndexedTriangleSet);
      sceneGraph.setX3dTextDoWithDataCallback(myText);
      sceneGraph.setX3dIndexedLineSetDoWithDataCallback(myIndexedLineSet);
      sceneGraph.setX3dPointSetDoWithDataCallback(myPointSet);
      sceneGraph.setX3dShapeDoWithDataCallback(myShape);
      sceneGraph.setX3dTransformDoWithDataCallback(myTransform);
      sceneGraph.setX3dGroupDoWithDataCallback(myGroup);
      sceneGraph.setX3dAnchorDoWithDataCallback(myAnchor);
      sceneGraph.setX3dSwitchDoWithDataCallback(mySwitch);
      sceneGraph.setX3dCollisionDoWithDataCallback(myCollision);
      sceneGraph.setX3dCADPartDoWithDataCallback(myCADPart);
      sceneGraph.setX3dCADAssemblyDoWithDataCallback(myCADAssembly);
      sceneGraph.setX3dCADLayerDoWithDataCallback(myCADLayer);
      sceneGraph.setX3dCADFaceDoWithDataCallback(myCADFace);
      sceneGraph.root.m_data = rootNode;
      sceneGraph.doWithData();
      LightState lightState = DisplaySystem.getDisplaySystem().getRenderer().createLightState();

      lightState.detachAll();

      DirectionalLight light = new DirectionalLight();
      light.setDirection(new Vector3f(0.0f, 0.0f, -1.0f));
      light.setDiffuse(new ColorRGBA(1.0f, 1.0f, 1.0f, 1.0f));
      light.setAmbient(new ColorRGBA(0.1f, 0.1f, 0.1f, 1.0f));
      light.setConstant(100);
      light.setEnabled(true);
      lightState.attach(light);

      DirectionalLight light2 = new DirectionalLight();
      light2.setDirection(new Vector3f(0.0f, 0.0f, 1.0f));
      light2.setDiffuse(new ColorRGBA(1.0f, 1.0f, 1.0f, 1.0f));
      light2.setAmbient(new ColorRGBA(0.1f, 0.1f, 0.1f, 1.0f));
      light2.setConstant(100);
      light2.setEnabled(true);
      lightState.attach(light2);

      rootNode.setRenderState(lightState);
   }
};
